const Container = require('./Container')

/** Manages services and injects classes from its service container. */
class DependencyInjector {
    constructor(container = new Container()) {
        /**
         * The service container used by this dependency injector.
         * @type {Container}
         */
        this.container = container
        this.container.di = this
    }

    /**
     * Inject a static class with the services it requests. This mutates
     * the class' prototype.
     * @param Class
     * @returns {*} - the injected static reference to the Class
     */
    make(Class) {
        if ( !this.__is_injectable(Class) ) {
            throw new Error('Cannot inject non-injectable class: '+Class.name)
        }

        Class.__inject(this.container)
        return Class
    }

    /**
     * Fetch a service by name. If no name is provided, return the service
     * proxy container. This container has getters for all the services by
     * name.
     * @param {string} service - the name of the service
     * @returns {{}|*} - the service instance of service container proxy
     */
    service(name) {
        return this.container.service(name)
    }

    /**
     * Verify that the injector's container has a service or set of services.
     * @param {string|string[]} name - service name or array of service names
     * @returns {boolean} - true if the container has the service(s)
     */
    has(name) {
        if ( Array.isArray(name) ) {
            return name.some(x => this.container.has(x))
        }

        return this.container.has(name)
    }

    /**
     * If called, this method will extend the global nodejs require() method to
     * check for injectable classes. If a require value is injectable, it will be
     * automatically injected with the services from this DI.
     */
    inject_globally() {
        const Module = require('module')
        const original_require = Module.prototype.require
        const di = this

        di._original_require = original_require

        Module.prototype.require = function() {
            const value = original_require.apply(this, arguments)
            return di.__is_injectable(value) ? di.make(value) : value
        }
    }

    /**
     * Verify that a class is injectable. This means that it has a static __inject
     * method and that method takes at least one argument. In almost all cases, this
     * should be satisfied by using the Injectable base class.
     * @param {*} Class - the class to check
     * @returns {boolean} - true if the class is injectable
     * @private
     */
    __is_injectable(Class) {
        return (
            typeof Class === 'function'
            && typeof Class.__inject === 'function'
            && Class.__inject.length > 0
        )
    }
}

module.exports = exports = DependencyInjector
